from typing import Any, Callable, Final, Iterator, TypeAlias

import numpy as np
from numpy.typing import NDArray

from ..base import BaseSolver

Robot: TypeAlias = tuple[int, int]
Grid: TypeAlias = list[list[int]]
ImageGrid: TypeAlias = NDArray[np.uint8]


class BoxUtils:
    FREE: Final = 0
    BLOCK: Final = 1

    @staticmethod
    def convert(grid_s: list[str], large: bool):
        grid: list[list[int]] = []
        robot: tuple[int, int] | None = None
        box_counter = 2 if not large else 3
        for i_row, row in enumerate(grid_s):
            row_u: list[int] = []
            for i_col, col in enumerate(row):
                if col == "." or col == "@":
                    row_u.extend(
                        (BoxUtils.FREE, BoxUtils.FREE) if large else (BoxUtils.FREE,)
                    )
                    robot = (i_row, i_col)
                elif col == "#":
                    row_u.extend(
                        (BoxUtils.BLOCK, BoxUtils.BLOCK) if large else (BoxUtils.BLOCK,)
                    )
                else:
                    row_u.extend(
                        (box_counter, -box_counter) if large else (box_counter,)
                    )
                    box_counter += 2
            grid.append(row_u)
        assert robot is not None
        return grid, robot

    @staticmethod
    def is_box(c: int):
        return c >= 2 and c % 2 == 0

    @staticmethod
    def is_open_or_close_box(c: int):
        return abs(c) >= 2 and c % 2 == 1

    @staticmethod
    def is_open_box(c: int):
        return c >= 2 and c % 2 == 1

    @staticmethod
    def is_close_box(c: int):
        return c < 0

    @staticmethod
    def to_char(c: int):
        if c == BoxUtils.FREE:
            return "."
        elif c == BoxUtils.BLOCK:
            return "#"
        elif BoxUtils.is_box(c):
            return "O"
        elif BoxUtils.is_open_box(c):
            return "["
        else:
            return "]"


class Solver(BaseSolver):
    def print_grid(self, name: str, grid: Grid, robot: Robot):
        if self.files:
            grid_s = [[BoxUtils.to_char(c) for c in row] for row in grid]
            grid_s[robot[0]][robot[1]] = "\033[31;1m@\033[00m"
            self.files.create(
                name, "\n".join("".join(row) for row in grid_s).encode(), True
            )

    def convert_grid(self, grid: Grid, robot: Robot) -> ImageGrid:
        import numpy as np

        # grid[robot[0]][robot[1]] = "@"
        # return np.array(
        #     [
        #         [
        #             # {
        #             #     "#": [255, 255, 255],
        #             #     "@": [255, 0, 0],
        #             #     "O": [0, 255, 0],
        #             #     "[": [0, 0, 255],
        #             #     "]": [0, 0, 255],
        #             #     ".": [0, 0, 0],
        #             # }[col]
        #             ord(col)
        #             for col in row
        #         ]
        #         for row in grid
        #     ],
        #     dtype=np.uint8,
        # )
        return np.abs(np.array(grid))

    def step_part1(self, grid: Grid, move: str, robot: Robot):
        match move:
            case "^":
                d_row, d_col = -1, 0
            case ">":
                d_row, d_col = 0, 1
            case "v":
                d_row, d_col = 1, 0
            case "<":
                d_row, d_col = 0, -1
            case _:
                assert False

        row, col = robot
        if grid[row + d_row][col + d_col] == BoxUtils.FREE:
            robot = (row + d_row, col + d_col)
        elif grid[row + d_row][col + d_col] != BoxUtils.BLOCK:
            n = 1
            while BoxUtils.is_box(grid[row + n * d_row][col + n * d_col]):
                n += 1

            if grid[row + n * d_row][col + n * d_col] == BoxUtils.FREE:
                robot = (row + d_row, col + d_col)
                for k in range(2, n + 1):
                    grid[row + k * d_row][col + k * d_col] = grid[
                        row + (k - 1) * d_row
                    ][col + (k - 1) * d_col]
                grid[row + d_row][col + d_col] = BoxUtils.FREE

        return grid, robot

    def step_part2(self, grid: Grid, move: str, robot: Robot):
        match move:
            case "^":
                d_row, d_col = -1, 0
            case ">":
                d_row, d_col = 0, 1
            case "v":
                d_row, d_col = 1, 0
            case "<":
                d_row, d_col = 0, -1
            case _:
                assert False

        row, col = robot
        if grid[row + d_row][col + d_col] == BoxUtils.FREE:
            robot = (row + d_row, col + d_col)
        elif grid[row + d_row][col + d_col] == BoxUtils.BLOCK:
            ...
        elif move in "<>":
            n = 1
            while BoxUtils.is_open_or_close_box(grid[row][col + n * d_col]):
                n += 1

            if grid[row][col + n * d_col] == BoxUtils.FREE:
                robot = (row, col + d_col)
                for k in range(n, 1, -1):
                    grid[row][col + k * d_col] = grid[row][col + (k - 1) * d_col]
                grid[row + d_row][col + d_col] = BoxUtils.FREE

        elif move in "^v":
            n = 1
            boxes: list[set[int]] = [{col}]
            while True:
                to_move = boxes[-1]
                if any(grid[row + n * d_row][c] == BoxUtils.BLOCK for c in to_move):
                    break
                if all(grid[row + n * d_row][c] == BoxUtils.FREE for c in to_move):
                    break

                as_move: set[int] = set()

                for c in to_move:
                    if BoxUtils.is_close_box(grid[row + n * d_row][c]):
                        as_move.update({c - 1, c})
                    elif BoxUtils.is_open_box(grid[row + n * d_row][c]):
                        as_move.update({c, c + 1})

                boxes.append(as_move)
                n += 1

            if all(grid[row + n * d_row][c] == BoxUtils.FREE for c in boxes[-1]):
                for k, to_move in zip(range(n, 1, -1), boxes[-1:0:-1], strict=True):
                    for c in to_move:
                        grid[row + k * d_row][c] = grid[row + (k - 1) * d_row][c]
                        grid[row + (k - 1) * d_row][c] = BoxUtils.FREE
                robot = (row + d_row, col + d_col)

        return grid, robot

    def run(
        self,
        name: str,
        grid: Grid,
        robot: Robot,
        moves: str,
        fn: Callable[
            [Grid, str, Robot],
            tuple[Grid, Robot],
        ],
        generate: bool,
    ) -> tuple[Grid, list[ImageGrid]]:
        # initialize
        images: list[ImageGrid] = []

        if generate:
            images.append(self.convert_grid(grid, robot))

        self.print_grid(f"initial_grid_{name}.txt", grid, robot)

        for move in self.progress.wrap(moves):
            self.logger.debug(f"Move '{move}'...")
            grid, robot = fn(grid, move, robot)

            if generate:
                images.append(self.convert_grid(grid, robot))

        self.print_grid(f"final_grid_{name}.txt", grid, robot)

        return grid, images

    def solve(self, input: str) -> Iterator[Any]:
        grid_s, moves = input.split("\n\n")
        moves = "".join(moves.split())

        grid, images = self.run(
            "part1",
            *BoxUtils.convert(grid_s.splitlines(), False),
            moves,
            self.step_part1,
            self.files is not None,
        )
        if self.files:
            print(images[0].max())
            self.files.image(
                "anim_part1.gif", np.stack(images, axis=0).astype(np.uint8)
            )
        yield sum(
            100 * i_row + i_col
            for i_row, row in enumerate(grid)
            for i_col, col in enumerate(row)
            if BoxUtils.is_box(col)
        )

        grid, images = self.run(
            "part2",
            *BoxUtils.convert(grid_s.splitlines(), True),
            moves,
            self.step_part2,
            self.files is not None,
        )
        if self.files:
            self.files.image(
                "anim_part2.gif", np.stack(images, axis=0).astype(np.uint8)
            )
        yield sum(
            100 * i_row + i_col
            for i_row, row in enumerate(grid)
            for i_col, col in enumerate(row)
            if BoxUtils.is_open_box(col)
        )
